@@ -83,10 +83,7 @@ bool DbConnection::with_alt_types() const {
 
 void DbConnection::set_busy_handler(SEXP r_callback) {
   check_connection();
-  if (busy_callback_) {
-    R_ReleaseObject(busy_callback_);
-    busy_callback_ = NULL;
-  }
+  release_callback_data();
 
   if (! Rf_isNull(r_callback)) {
     R_PreserveObject(r_callback);
@@ -107,24 +104,38 @@ void DbConnection::release_callback_data() {
   }
 }
 
-int DbConnection::busy_callback_helper(void *data, int num) {
-  SEXP r_callback = reinterpret_cast <SEXP> (data);
+int DbConnection::busy_callback_helper(void *data, int num)
+{
+  SEXP r_callback = reinterpret_cast<SEXP>(data);
 
-  try {
-    Function rfun = r_callback;
-    IntegerVector ret = rfun(num);
-    return as<int>(ret);
-
-  } catch(eval_error &e) {
-    Rcpp::warning("Busy callback failed, aborting transaction: %s", e.what());
-    return 0;
-
-  } catch(Rcpp::internal::InterruptedException &e) {
-    // Not warning on explicit interrupt
-    return 0;
-
-  } catch(...) {
-    Rcpp::warning("Busy callback failed, aborting transaction");
+  // Overarching safety net
+  try
+  {
+    try
+    {
+      Function rfun = r_callback;
+      IntegerVector ret = rfun(num);
+      return as<int>(ret);
+    }
+    catch (eval_error &e)
+    {
+      std::string msg = std::string("Busy callback failed, aborting transaction: ") + e.what();
+      Rcpp::message(Rcpp::StringVector::create(msg));
+      return 0;
+    }
+    catch (Rcpp::internal::InterruptedException &e)
+    {
+      // Not warning on explicit interrupt
+      return 0;
+    }
+    catch (...)
+    {
+      Rcpp::message(Rcpp::StringVector::create("Busy callback failed, aborting transaction"));
+      return 0;
+    }
+  }
+  catch (...)
+  {
     return 0;
   }
 }
